#include "rasteriser/rasteriser.h"
#include "c_cpp_aliases/aliases.h"
#include "list/typed_list.h"
#include "vector/vec.h"
#include "window/window.h"
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>

internal list_float_t *interpolate(i32 i0, f32 d0, i32 i1, f32 d1);

internal inline void order_triangle_points(triangle_t *triangle);

void draw_wireframe_triangle(window_t *wnd, triangle_t triangle,
                             colour_t colour) {
  order_triangle_points(&triangle);

  draw_line(wnd, (line_t){triangle.p0, triangle.p1}, colour);
  draw_line(wnd, (line_t){triangle.p1, triangle.p2}, colour);
  draw_line(wnd, (line_t){triangle.p2, triangle.p0}, colour);
}

void draw_filled_triangle(window_t *wnd, triangle_t triangle, colour_t colour) {
  order_triangle_points(&triangle);

  i32 x0 = triangle.p0.x;
  i32 y0 = triangle.p0.y;
  i32 x1 = triangle.p1.x;
  i32 y1 = triangle.p1.y;
  i32 x2 = triangle.p2.x;
  i32 y2 = triangle.p2.y;

  list_float_t *x01 = interpolate(y0, x0, y1, x1);
  list_float_t *x12 = interpolate(y1, x1, y2, x2);
  list_float_t *x02 = interpolate(y0, x0, y2, x2);
  list_float_t *x012 = NULL;

  list_pop(x01); // Last element of x01 is a duplicate of first element in x12
  list_merge(f32, x012, x01, x12);

  list_float_t *x_left;
  list_float_t *x_right;
  u64 middle = (u64)(floorf((f32)(x02->count) / 2.0f));
  if (list_get(x02, middle) < list_get(x012, middle)) {
    x_left = x02;
    x_right = x012;
  } else {
    x_left = x012;
    x_right = x02;
  }

  for (i64 y = y0; y <= y2; ++y) {
    for (i64 x = list_get(x_left, y - y0); x < list_get(x_right, y - y0); ++x) {
      set_pixel(wnd, x, y, colour);
    }
  }
}

void draw_line(window_t *wnd, line_t line, colour_t colour) {
  list_float_t *values = NULL;

  if (abs(line.p1.x - line.p0.x) > abs(line.p1.y - line.p0.y)) {
    if (line.p1.x < line.p0.x) {
      vec_swap(vec2i_t, line.p0, line.p1);
    }

    i32 x0 = line.p0.x;
    i32 y0 = line.p0.y;
    i32 x1 = line.p1.x;
    i32 y1 = line.p1.y;

    values = interpolate(x0, y0, x1, y1);
    if (!values) {
      return;
    }

    for (i32 x = x0; x <= x1; ++x) {
      set_pixel(wnd, x, (i32)(list_get(values, x - x0)), colour);
    }
  } else {
    if (line.p1.y < line.p0.y) {
      vec_swap(vec2i_t, line.p0, line.p1);
    }

    i32 x0 = line.p0.x;
    i32 y0 = line.p0.y;
    i32 x1 = line.p1.x;
    i32 y1 = line.p1.y;

    values = interpolate(y0, x0, y1, x1);
    if (!values) {
      return;
    }

    for (i32 y = y0; y <= y1; ++y) {
      set_pixel(wnd, (i32)(list_get(values, y - y0)), y, colour);
    }
  }

  list_destroy(f32, values);
}

internal list_float_t *interpolate(i32 i0, f32 d0, i32 i1, f32 d1) {
  list_float_t *values;
  if (i0 == i1) {
    values = list_create_with_capacity(f32, 20);
    if (values) {
      list_append(f32, values, d0);
    }

    return values;
  }

  values = list_create(f32);
  if (!values) {
    return NULL;
  }

  f32 a = (d1 - d0) / ((f32)i1 - i0);
  f32 d = d0;

  for (i32 i = i0; i <= i1; ++i) {
    list_append(f32, values, d);
    d += a;
  }

  return values;
}

internal inline void order_triangle_points(triangle_t *triangle) {
  if (triangle->p1.y < triangle->p0.y) {
    vec_swap(vec2i_t, triangle->p0, triangle->p1);
  }
  if (triangle->p2.y < triangle->p0.y) {
    vec_swap(vec2i_t, triangle->p0, triangle->p2);
  }
  if (triangle->p2.y < triangle->p1.y) {
    vec_swap(vec2i_t, triangle->p1, triangle->p2);
  }
}
